
# 単純メモリbackend (SLDB-backend) [memory_backend]

import os
import sys
import fies
import atexit
import pickle
from sout import sout, souts

# プログラム終了時にcommitを実行する
commit_target_ls = []	# コミット対象
def cleanup():
	for one_backend in commit_target_ls:
		one_backend.commit()
# プログラム終了時に呼び出す関数を登録
atexit.register(cleanup)

# 初期化されていない場合は初期化
def init_if_needed(data_filename, fmt):
	if os.path.exists(data_filename) is True: return None
	fies[data_filename, fmt] = {}

# メモリバックエンドのクラス
class Memory_Backend:
	# 初期化処理
	def __init__(self, data_dir,
		fmt = "fpkl",	# バックエンドのファイル形式 (fpkl: fast-pickle(default), pickle: pickle)
	):
		# バックエンドのファイル形式 (fpkl: fast-pickle(default), pickle: pickle)
		self.fmt = fmt
		# データ保存パス
		self.data_filename = os.path.join(data_dir, f"memory_backend_data.{self.fmt}")
		# 初期化されていない場合は初期化
		init_if_needed(self.data_filename, self.fmt)
		# ファイルの内容をメモリに展開
		self.data = fies[self.data_filename, self.fmt]
		# プログラム終了時のcommit対象に自身を追加
		commit_target_ls.append(self)
	# create, update共通 [memory_backend]
	def __setitem__(self, key, value): self.data[key] = value
	# key読み出し [memory_backend]
	def __getitem__(self, key): return self.data[key]
	# 削除 [memory_backend]
	def __delitem__(self, key): del self.data[key]
	# 存在確認 [memory_backend]
	def __contains__(self, key): return (key in self.data)
	# 強制commit [memory_backend]
	def commit(self):
		fies[self.data_filename, self.fmt] = self.data
	# 文字列化 (その1, その2)
	def __str__(self): return f"<SLDB-memory_backend {souts(self.data)}>"
	def __repr__(self): return str(self)
	# 非公式API: イテレーション
	def __iter__(self): return iter(self.data)
	# 非公式API: 要素数取得
	def __len__(self): return len(self.data)

# メモリバックエンドへの接続 [memory_backend]
def conn(
	data_dir,	# バックエンドの情報を記録するディレクトリ
	fmt = "fpkl",	# バックエンドのファイル形式 (fpkl: fast-pickle(default), pickle: pickle)
):
	# メモリバックエンドのクラス
	return Memory_Backend(data_dir, fmt = fmt)




# 双方向グラフDB [SymLinkDB]
# 【動作確認 / 使用例】

import os
import sys
import shutil
from tqdm import tqdm
from sout import sout
from ezpip import load_develop
# 双方向グラフDB [SymLinkDB]
SymLinkDB = load_develop("SymLinkDB", "../", develop_flag = True)

# データベースディレクトリの準備
db_dir = "./document_test"
if not os.path.exists(db_dir):
    os.makedirs(db_dir)

# バックエンドの初期化
backend = SymLinkDB.memory_backend(db_dir)  # デフォルトの単純なバックエンド

# データベース接続の確立
sldb = SymLinkDB.conn(backend=backend)

# テーブルのロード
sldb.load_table("user_table")
sldb.load_table("tool_table")

# リンクのロード
sldb.load_link(
    "所有関係",
    ("user_table", "所有者", "1"),
    ("tool_table", "所有物", "N")
)

# user_tableにレコードを追加
user_rec = sldb["user_table"].create(data={"name": "Alice"})

# tool_tableにレコードを追加し、所有関係を設定
tool_rec = sldb["tool_table"].create(data="Hammer", links={("所有関係", "所有者"): user_rec})

# 逆向きのリンクも自動的に追加されている
print(tool_rec in user_rec["所有物"])  # -> True

# レコードのデータを更新
user_rec.data = {"name": "Alice", "age": 30}

# 特定のレコードをIDを使用して取得
retrieved_rec = sldb["user_table"][user_rec.id]

# レコードのデータを読み出し
retrieved_data = retrieved_rec.data
print(retrieved_data)   # -> {'name': 'Alice', 'age': 30}

# del sldb["user_table"][user_rec]

# 新しいリンクを追加
new_tool_rec = sldb["tool_table"].create(data="Screwdriver")
link_set = user_rec["所有物"]    # 対象レコードに関係するリンク群 (link_set) を取得
# link_set = user_rec[("所有関係", "所有物")]    # role名のみでは指定が曖昧になる場合は、このようにタプルを用いた正式な参照を用いる
link_set.push(new_tool_rec)

# link_set内の要素をイテレート [SymLinkDB]
for rec in link_set: print(rec) # リンク先のレコードが順次取得される

flag = (new_tool_rec in link_set)    # linkの存在確認 [SymLinkDB]
print(len(link_set))  # link_set内の要素数を取得 [SymLinkDB]

# リンクを削除
del link_set[new_tool_rec]




# 双方向グラフDB [SymLinkDB]
# 【動作確認 / 使用例】

import os
import sys
import shutil
from tqdm import tqdm
from sout import sout
from ezpip import load_develop
# 双方向グラフDB [SymLinkDB]
SymLinkDB = load_develop("SymLinkDB", "../", develop_flag = True)

# 例外のassert (「例外が発生し、所定のキーワードを含むこと」)
def assert_err(
	proc_func,	# 対象の処理 (ゼロ引数関数)
	msg = ""	# 例外が含むべき文字列 (省略された場合は、「例外は発生すべきだが、文言は問わない」)
):
	try:
		proc_func()
	except Exception as err:
		assert (msg in repr(err))
		return True
	# 正常に処理が完了した場合はassert失敗
	raise Exception("[assert_err error] 例外が発生すべき処理について正常終了したため、想定外です")

# --- テスト準備 ---

# テスト用にディレクトリを削除・再作成
if os.path.exists("./db_dir"): shutil.rmtree("./db_dir")
for dirname in [f"./db_dir/{i+1}/" for i in range(5)]:
	os.makedirs(dirname)	# 消し忘れの場合等はわざとエラーにする

# バックエンドを初期化
m_back1 = SymLinkDB.memory_backend("./db_dir/1/")	# メモリバックエンドの初期化 [SymLinkDB]
m_back2 = SymLinkDB.memory_backend("./db_dir/2/")	# メモリバックエンドの初期化 [SymLinkDB]
m_back3 = SymLinkDB.memory_backend("./db_dir/3/")	# メモリバックエンドの初期化 [SymLinkDB]
m_back4 = SymLinkDB.memory_backend("./db_dir/4/")	# メモリバックエンドの初期化 [SymLinkDB]

# SLDB初期化 [SymLinkDB]
sldb = SymLinkDB.conn(backend = m_back1)

# tableの読み込み (存在しない場合は空で初期化される) [SymLinkDB]
sldb.load_table("user_table", search_keys = ["name", "age"], backend = m_back2)
# tableの読み込み (存在しない場合は空で初期化される) [SymLinkDB]
sldb.load_table("tool_table", backend = m_back3)
# linkの読み込み (存在しない場合は空で初期化される) [SymLinkDB]
sldb.load_link(
	"所有関係",	# テーブル名
	("user_table", "所有者", "1"),	# 関係性情報0 (table名, role, 1 or N)
	("tool_table", "所有物", "N"),	# 関係性情報1 (table名, role, 1 or N)
	backend = m_back4
)

# --- SLDBのテスト ---

# 文字列化 (その1, その2)
sldb_str = str(sldb)	# -> '<SymLinkDB tables = ["user_table", "tool_table"]>'
print(sldb_str)
assert sldb_str == '<SymLinkDB tables = ["user_table", "tool_table"]>'

assert len(sldb) == 2	# table数取得 [SymLinkDB]
assert ("user_table" in sldb)	# tableの存在確認 [SymLinkDB]
assert ("dummy_table_name" not in sldb)	# tableの存在確認 [SymLinkDB]

assert list(sldb) == ['user_table', 'tool_table']	# tableのイテレート [SymLinkDB]

# --- tableのテスト ---

# tableの取得 [SymLinkDB]
tool_table = sldb["tool_table"]
# レコードの新規作成 [SymLinkDB]
tool1_rec = tool_table.create(data = "tool1", links = {("所有関係", "所有者"): None})
print(f"tool1_id: {tool1_rec.id}")
tool2_rec = tool_table.create(data = "tool2", links = {"所有者": None})
tool3_rec = tool_table.create(data = "tool3", links = {"所有者": []})
# 文字列化 (その1, その2)
table_str = str(tool_table)	# -> '<SymLinkTable tool_table recs = {"dIMcC": <Rec "tool1">, ... (n = 3)}>'
print(table_str)
assert "<SymLinkTable tool_table recs = " in table_str
assert "(n = 3)" in table_str

user_table = sldb["user_table"]	# tableの取得 [SymLinkDB]

user1_rec = user_table.create(data = {"name": "user1", "age": 1}, links = {"所有物": [tool1_rec, tool2_rec]})	# レコードの新規作成 [SymLinkDB]
user2_rec = user_table.create(data = {"name": "user2", "age": {"4":"5", "2":"1"}}, links = {"所有物": None})	# レコードの新規作成 [SymLinkDB]
user3_rec = user_table.create(data = {"name": "user3", "age": {4:5, 2:1}}, links = {})	# レコードの新規作成 [SymLinkDB]
user3_id = user3_rec.id
del user_table[user3_rec]	# レコードの削除 [SymLinkDB]
assert_err(lambda: user3_rec["所有物"], msg = "This record is already deleted.")	# 例外のassert (「例外が発生し、所定のキーワードを含むこと」)
assert user1_rec in user_table	# レコードの存在確認 [SymLinkDB]
assert user1_rec.id in user_table
assert user3_rec.id not in user_table
assert len(user_table) == 2	# レコード数取得 [SymLinkDB]
# tableのイテレート [SymLinkDB]
assert set([rec.id for rec in user_table]) == {user1_rec.id, user2_rec.id}

# レコードの取得 [SymLinkDB]
got_user1_rec = user_table[user1_rec.id]
assert user1_rec.id == got_user1_rec.id

print(user_table[{"name": "user", "age": 2}])

# --- recordのテスト ---

# レコードの文字列化 [SymLinkDB]
rec_str = str(user1_rec)	# -> '<SymLinkRecord user_table.(レコードのID) data = {"name": "user1"} links = [("所有関係", "所有物")]>'
print(rec_str)
assert rec_str == f'<SymLinkRecord user_table.{user1_rec.id} data = {{"name": "user1", "age": 1}} links = [("所有関係", "所有物")]>'

# 削除済みレコードの文字列化
assert str(user3_rec) == "<SymLinkRecord (deleted)>"

user1_rec.data = {"name": "taro", "age": 23}	# レコードのdataのupdate [SymLinkDB]
assert user1_rec.data == {'name': 'taro', 'age': 23}	# レコードのdataの取得 [SymLinkDB]
print(user1_rec.id)	# レコードのidの取得 [SymLinkDB]
assert user1_rec.table_name == "user_table"	# レコードの所属table名の取得 [SymLinkDB]
table = user1_rec.table	# レコードの所属tableオブジェクトの取得 [SymLinkDB]
assert table.table_id == user_table.table_id
assert set([key for key in user1_rec]) == {('所有関係', '所有物')}	# レコードのイテレート [SymLinkDB]
assert ('所有関係', '所有物') in user1_rec	# 指定されたlinkの存在確認 [SymLinkDB]
assert '所有物' in user1_rec	# 指定されたlinkの存在確認 [SymLinkDB]

# --- link_setのテスト ---

# link_setの取得 [SymLinkDB]
user1_link_set = user1_rec[("所有関係", "所有物")]

# 略指定による取得
assert {r.id for r in user1_rec["所有物"]} == {r.id for r in user1_link_set}

# linkの追加 [SymLinkDB]
user1_link_set.push(tool3_rec)

# linkの削除 [SymLinkDB]
del user1_link_set[tool2_rec]

# linkの存在確認 [SymLinkDB]
assert tool3_rec in user1_link_set
assert tool3_rec.id in user1_link_set	# レコードIDでも存在確認可能
assert tool2_rec not in user1_link_set

# link_set内の要素をイテレート [SymLinkDB]
assert set([rec.id for rec in user1_link_set]) == {tool1_rec.id, tool3_rec.id}

assert len(user1_link_set) == 2	# link_set内の要素数を取得 [SymLinkDB]

# 文字列化 (その1, その2)
u1_l_set_str = str(user1_link_set)	# -> "<LinkSet 所有関係: user_table(所有者;1) -> tool_table(所有物;N;n=2)>"
print(u1_l_set_str)
assert u1_l_set_str == "<LinkSet 所有関係: user_table(所有者;1) -> tool_table(所有物;N;n=2)>"

# 強制コミット [SymLinkDB]
sldb.commit()

# --- 発展的な使い方 ---

# CachedFileDicバックエンドの利用
cfd_backend = SymLinkDB.cfd_backend("./db_dir/5/")	# CachedFileDicバックエンドの初期化 [SymLinkDB]
sldb.load_table("example_table", backend = cfd_backend)	# tableの読み込み (存在しない場合は空で初期化される) [SymLinkDB]

# dataの型は任意
rec = tool_table.create(data = b"binary_data")
print(rec)	# -> "<SymLinkRecord user_table.(レコードのID) data = b'binary_data' links = [("所有関係", "所有物")]>"

# バックエンドを明示的に指定しない場合は、SymLinkDB.conn()時のbackendが使われる
sldb.load_table("pictures_table")	# tableの読み込み (存在しない場合は空で初期化される) [SymLinkDB]

# --- 異常系や複雑な場合のテスト ---

# 1-N関係を無視したpush
def err_push_1(): tool1_rec["所有者"].push(user2_rec)
assert_err(err_push_1, "Violation of 1-N constraint")

# 既にあるIDのpush
def err_push_2(): tool1_rec["所有者"].push(user1_rec)
assert_err(err_push_2, "This is a link that has already been added")

# 全く無関係なIDのpush
def err_push_3(): tool1_rec["所有者"].push(tool2_rec.id)
assert_err(err_push_3, "specified record ID does not exist in the table")

# 複数同時delete, push
del user1_link_set[tool1_rec, tool3_rec.id]
user1_link_set.push([tool1_rec, tool2_rec.id])
assert {rec.id for rec in user1_link_set} == {tool1_rec.id, tool2_rec.id}

# link両端のrole名が被っているときに例外を出す
def err_link_1():
	sldb.load_link(	# linkの読み込み (存在しない場合は空で初期化される) [SymLinkDB]
		"友人関係",	# テーブル名
		("user_table", "人", "1"),	# 関係性情報0 (table名, role, 1 or N)
		("user_table", "人", "1"),	# 関係性情報1 (table名, role, 1 or N)
	)
assert_err(err_link_1, "The role names at both ends of the link must be different.")

# 1-table内関係・N-N関係
sldb.load_link(	# linkの読み込み (存在しない場合は空で初期化される) [SymLinkDB]
	"親子関係",	# テーブル名
	("user_table", "親", "N"),	# 関係性情報0 (table名, role, 1 or N)
	("user_table", "子", "N"),	# 関係性情報1 (table名, role, 1 or N)
)

child_link_set_1 = user1_rec["子"]
user4_rec = user_table.create({"name": "user4", "age": 4})
user5_rec = user_table.create({"name": "user5", "age": 4})
child_link_set_1.push([user2_rec, user4_rec])
user4_rec["親"].push(user5_rec)
user4_rec["子"].push(user1_rec)
del user4_rec["子"][user1_rec]
assert {rec.id for rec in user2_rec["親"]} == {user1_rec.id}
assert {rec.id for rec in user4_rec["親"]} == {user1_rec.id, user5_rec.id}
assert {rec.id for rec in user5_rec["子"]} == {user4_rec.id}
assert {rec.id for rec in user1_rec["子"]} == {user2_rec.id, user4_rec.id}
assert {rec.id for rec in user1_rec["親"]} == set([])	# deleteの動作確認
assert {rec.id for rec in user2_rec["子"]} == set([])	# 勝手に逆向きリンクが加わらないことのテスト

# table - del済みレコードのdel
def err_del_1():
	del user_table[user3_id]	# レコードの削除 [SymLinkDB]
assert_err(err_del_1, "You cannot delete a record that does not exist")

# delete時のlink自動削除
del tool_table[tool2_rec]
assert tool2_rec not in user1_rec["所有物"]

print("TEST DONE")
sys.exit()

# rec数の大きなテーブルでCRUD速度テスト
large_n = 100000
sldb.load_table("large_data_test_table")
ldt_table = sldb["large_data_test_table"]
print("creating...")
for i in tqdm(range(large_n)):
	ldt_table.create(data = {"hoge": f"data_{i%10}" * (10000 // 6)})
print("reading and updating...")
for rec in tqdm(ldt_table):
	now_data = rec.data
	rec.data = {**now_data, "fuga": "new_data"}
print('running "contains"...')
for rec in tqdm(ldt_table): (rec in ldt_table)
print('reading ids...')
rec_id_ls = [rec.id for rec in tqdm(ldt_table)]
print('deleting...')
for rec in tqdm(rec_id_ls):
	del ldt_table[rec]

# --- テスト用資材

# # カニ差分テスト
# def tr(dic_like): return {k: dic_like[k] for k in dic_like}
# from sout import souts
# import fies
# fies["debug_pre.json", "t"] = souts(tr(link_table_backend), None)
# # 何らかの処理
# fies["debug_post.json", "t"] = souts(tr(link_table_backend), None)
# print("done!")
# sys.exit()




# 双方向グラフDB [SymLinkDB]

import os
import sys
import fies
import json
import atexit
import pickle
import slim_id
from sout import sout, souts

# 十分な長さの64進idの発行 (uuidの桁数を参考に決定)
def gen_unique_id(): return slim_id.gen(lambda e: False, length = 22)

# プログラム終了時にcommitを実行する
commit_target_dbs = []	# コミット対象
def cleanup():
	for sldb in commit_target_dbs:
		sldb.commit()	# 強制コミット [SymLinkDB]
# プログラム終了時に呼び出す関数を登録
atexit.register(cleanup)

# メモリバックエンドの初期化 [SymLinkDB]
def memory_backend(backend_dir):
	# 単純メモリbackend (SLDB-backend) [memory_backend]
	from .parts import memory_backend as m_back_module
	# メモリバックエンドへの接続 [memory_backend]
	return m_back_module.conn(backend_dir)

# CachedFileDicバックエンドの初期化 [SymLinkDB]
def cfd_backend(backend_dir):
	# CachedFileDic-backend (SLDB-backend) [cfd_backend]
	from .parts import cfd_backend as cfd_back_module
	# CachedFileDicバックエンドへの接続 [cfd_backend]
	return cfd_back_module.conn(backend_dir)

# SLDB-backendとして初期化
def init_sldb_backend(backend):
	# このバックエンドを一意に特定するID
	backend[("meta", "backend_id")] = gen_unique_id()	# 十分な長さの64進idの発行 (uuidの桁数を参考に決定)
	# このバックエンドに格納されているデータの一覧
	backend[("meta", "backend_contain_data")] = []

# このバックエンドがSLDB-sysとして初期化されているかを判定
def judge_sys_initialized(backend):
	bcd = backend[("meta", "backend_contain_data")]
	# backend_contain_data 内に定義されたsysの数を数える
	sys_n = len([e for e in bcd
		if e["entity_type"] == "SLDB-system"])
	# 数に応じて分岐
	if sys_n == 0: return False
	if sys_n == 1: return True
	raise Exception("[SymLinkDB error] Inconsistency in database structure. (Despite the fact that this version of SymLinkDB can only have one system per backend, multiple systems are defined.)")

# 辞書-likeオブジェクトの所定のkeyのvalue(list)にデータを追加
def back_ls_push(
	target_dic_like,	# 追記対象の辞書-likeオブジェクト
	key,	# 編集対象のkey
	value	# 追記するデータ
):
	target_ls = target_dic_like[key]
	target_ls.append(value)
	target_dic_like[key] = target_ls

# このバックエンドをSLDB-sysとして初期化
def init_sldb_sys(backend):
	# systemの情報を記入
	backend[("meta", "system_contain_data")] = []
	# backend_contain_dataにsysを登録
	back_ls_push(	# 辞書-likeオブジェクトの所定のkeyのvalue(list)にデータを追加
		backend, ("meta", "backend_contain_data"),
		{"entity_type": "SLDB-system"})

# metaのcontain_dataを条件でフィルタリングする
def entity_filter(
	entity_ls,	# entityがたくさん入ったlist (metaのxx_contain_dataなど)
	cond,	# フィルタリング条件
	unique = False	# 検索結果が唯一でないといけないか (True指定の場合は戻り値がentityのリストではなくentityになる)
):
	class VOID: pass	# 他のどの値にも一致しない値 (等価性演算子に対して)
	def check_one(entity):
		for k in cond:
			if entity.get(k, VOID) != cond[k]: return False
		return True
	ret_ls = [e for e in entity_ls
		if check_one(e) is True]
	# uniqueかどうかで返し方を変える
	if unique is True:
		if len(ret_ls) != 1: raise Exception("[SymLinkDB error] The search results of entity_filter() violate the unique constraint.")
		return ret_ls[0]
	else:
		return ret_ls

# あるエンティティについて、バックエンドが整合しているかを確かめる (NGの場合は例外を送出する)
def check_entity_backend(
	filter_condition,	# エンティティのフィルタリング条件 (uniqueになるようなもの)
	sys_backend,	# DB全体情報 (system) が保管されているバックエンド
	table_data_backend,	# テーブルのデータが保管されているバックエンド
):
	# データが格納されているbackend-idが同じかを見る (system側とdata側の記載の整合性)
	sys_entity = entity_filter(sys_backend[("meta", "system_contain_data")],	# metaのcontain_dataを条件でフィルタリングする
		filter_condition, unique = True)
	b_id_A = sys_entity["data_backend_id"]
	b_id_B = (table_data_backend[("meta", "backend_id")]
		if (("meta", "backend_id") in table_data_backend) else None)	# backendが未初期化の場合も想定
	if b_id_A == b_id_B: return True
	raise Exception(f"[SymLinkDB error] The entityfilter_condition is already initialized, but there is no data in the specified backend. Please specify the correct backend. (Correct backend ID = {b_id_A})")

# search_keysの整合性を確認
def check_search_keys(
	search_keys,	# チェック対象のsearch_keys
	table_name,	# テーブル名
	system_contain_data,	# 該当sysの内容物一覧
):
	# 該当テーブルの情報を取得
	tbl_entity = entity_filter(system_contain_data,	# metaのcontain_dataを条件でフィルタリングする
		{"entity_type": "SLDB-table", "table_name": table_name}, unique = True)
	# 整合性違反の場合は例外を送出する
	ls_2_set = lambda e: (set(e) if type(e) == type([]) else e)
	a = ls_2_set(search_keys)
	b = ls_2_set(tbl_entity["search_keys"])
	if a != b: raise Exception("[SymLinkDB error] It differs from the search_keys defined initially.")

# テーブル新規作成処理
def create_table(
	table_name,	# テーブル名
	search_keys,	# 検索用に使いたい列名
	sys_backend,	# DB全体情報 (system) を保管するバックエンド
	table_data_backend,	# テーブルのデータを保管するバックエンド
):
	# table初期化不整合チェック
	filtered_entities = entity_filter(table_data_backend[("meta", "backend_contain_data")],
		{"table_name": table_name, "sys_backend_id": sys_backend[("meta", "backend_id")]})
	if len(filtered_entities) > 0: raise Exception("[SymLinkDB error] Inconsistency in the stored table information. According to the information on the DB-system side, the table that is currently being created is uninitialized, but within the backend that stores the table, it is already initialized. Degradation in material management is suspected.")
	# search_keysの形式チェック
	if search_keys is not None:
		if type(search_keys) != type([]): raise Exception("[SymLinkDB error] Please specify search_key in the form of a list of strings.")
		for e in search_keys:
			if type(e) != type(""): raise Exception("[SymLinkDB error] Please specify search_key in the form of a list of strings.")
	# table情報
	table_data = {
		"entity_type": "SLDB-table",
		"table_id": gen_unique_id(),	# 十分な長さの64進idの発行 (uuidの桁数を参考に決定)
		"sys_backend_id": sys_backend[("meta", "backend_id")],
		"data_backend_id": table_data_backend[("meta", "backend_id")],
		"table_name": table_name,
		"search_keys": search_keys,
	}
	# systemにtable情報を書き込む
	back_ls_push(	# 辞書-likeオブジェクトの所定のkeyのvalue(list)にデータを追加
		sys_backend, ("meta", "system_contain_data"),
		table_data)
	# backendにtable情報を書き込む
	back_ls_push(	# 辞書-likeオブジェクトの所定のkeyのvalue(list)にデータを追加
		table_data_backend, ("meta", "backend_contain_data"),
		table_data)
	# レコード数・レコード一覧を書き込む
	table_data_backend[("table_info", "all_rec_ids", table_data["table_id"])] = {}	# レコード数0
	table_data_backend[("table_info", "rec_n", table_data["table_id"])] = 0	# レコード数0

# 辞書-likeオブジェクトの所定keyのvalueを編集
def dic_like_edit(
	dic_like_obj, key,
	edit_func	# 値を編集する関数 (今の値を受け取って更新後の値を返す)
):
	dic_like_obj[key] = edit_func(dic_like_obj[key])

# レコード一覧を表した要約文字列の生成
def gen_recs_str(table):
	for rec in table: break	# table内のrec_idの一例を取り出す処理
	if len(table) == 0:
		inner = ""
	elif len(table) == 1:
		inner = f'"{rec.id}": <Rec {souts(rec.data)}>'
	else:
		inner = f'"{rec.id}": <Rec {souts(rec.data)}>, ... (n = {len(table)})'
	return "{"+ inner +"}"

# role指定からlink_table_nameを特定 (role指定が曖昧な場合は例外を投げる)
def role_to_link_name(
	obj_role,	# 相手のrole
	subj_table_id,	# 自分のtable_id
	sys_backend	# システムバックエンド
):
	# link_tableのentityのみに絞り込み
	lt_entities = entity_filter(sys_backend[("meta", "system_contain_data")],
		{"entity_type": "SLDB-link-table"})
	# roleとtable_idで絞り込み (両側見る)
	filtered_tn_ls = []
	for entity in lt_entities:
		role_0, role_1 = entity["rel_table_info"]
		for subj_dic, obj_dic in [(role_0, role_1), (role_1, role_0)]:
			# 一致しないものは除外
			if subj_dic["table_id"] != subj_table_id: continue
			if obj_dic["role"] != obj_role: continue
			# 候補の格納
			filtered_tn_ls.append(entity["link_table_name"])
	# 候補が唯一かどうかを確認して返す
	if len(filtered_tn_ls) > 1: raise Exception("[SymLinkDB error] Ambiguous link identification information is specified, and the link cannot be uniquely determined. If you have not specified the name of the link table, please consider specifying it.")
	if len(filtered_tn_ls) == 0: raise Exception("[SymLinkDB error] The role specification is invalid. There is no corresponding link.")
	link_table_name = filtered_tn_ls[0]
	return link_table_name

# role, link_table_name からlink特定情報を取得する
def get_link_info_core(link_table_name, role, table_id, sys_backend):
	# link_table_nameで絞り込み
	entity = entity_filter(sys_backend[("meta", "system_contain_data")],
		{"entity_type": "SLDB-link-table", "sys_backend_id": sys_backend[("meta", "backend_id")], "link_table_name": link_table_name},
		unique = True)
	# 対象tableがどちら側かを検索 (両側同一tableである可能性があることに注意)
	role_0, role_1 = entity["rel_table_info"]
	suspect_role = None	# 書き間違い疑いのrole名
	for subj_role, obj_role in [(role_0, role_1), (role_1, role_0)]:
		# 一致を確認
		if subj_role["table_id"] != table_id: continue
		if obj_role["role"] != role:
			suspect_role = obj_role["role"]
			continue
		# 情報を返す
		return {
			"link_table_id": entity["link_table_id"],
			"link_table_name": entity["link_table_name"],
			"subj_role_info": subj_role, "obj_role_info": obj_role,
		}
	# 不整合エラー
	if suspect_role is not None:
		raise Exception(f'[SymLinkDB error] The specified role name does not match. ("{suspect_role}" is suspected to be the correct role name)')
	raise Exception("[SymLinkDB error] The specified link_table_name does not have a relationship with the specified table.")

# role名あるいはrole名とlink table名からlink情報を取得
def get_link_info(query, table_id, system_backend):
	# roleのみで指定している場合 -> 完全な指定に変換
	if type(query) == type(""):
		role = query
		link_table_name = role_to_link_name(obj_role = role, subj_table_id = table_id, sys_backend = system_backend)	# role指定からlink_table_nameを特定 (role指定が曖昧な場合は例外を投げる)
		query = (link_table_name, role)	# 完全な指定に変換
	# 「完全な指定」の形式をチェック
	if type(query) != type(tuple()) or len(query) != 2: raise Exception("[SymLinkDB error] invalid type")
	# 「完全な指定」によるlink特定情報の取得
	link_table_name, role = query
	return get_link_info_core(link_table_name, role, table_id, system_backend)	# role, link_table_name からlink特定情報を取得する

# 1-Nチェック (片側; 制約違反の場合は例外を投げる)
def check_1N(
	subj,	# 「自分側」の情報 (rec_idとrole_info3つ)
	obj,	# 「相手側」の情報 (rec_idとrole_info3つ)
	link_table_backend,	# link tableの入っているバックエンド
	link_table_id	# link table ID
):
	# Nの場合はどんな場合もOK
	if obj["1-N"] == "N": return True
	# 1Nルール指定が不正な場合
	if obj["1-N"] != "1": raise Exception('[SymLinkDB error] Incorrect rule notation: Please specify either "1" or "N"')
	# 1の場合のチェック
	key = ("link_rec", link_table_id, subj["role"], obj["role"], subj["table_id"], subj["rec_id"])
	now_n = (len(link_table_backend[key]) if key in link_table_backend else 0)
	if now_n + 1 > 1: raise Exception("[SymLinkDB error] Violation of 1-N constraint")

# 片側のリンク追加
def add_link(
	subj,	# 「自分側」の情報 (rec_idとrole_info3つ)
	obj,	# 「相手側」の情報 (rec_idとrole_info3つ)
	link_table_backend,	# link tableの入っているバックエンド
	link_table_id,	# link table ID
):
	key = ("link_rec", link_table_id, subj["role"], obj["role"], subj["table_id"], subj["rec_id"])
	# keyがまだ存在しない場合は空のレコードを追加
	if key not in link_table_backend: link_table_backend[key] = []
	# 辞書-likeオブジェクトの所定のkeyのvalue(list)にデータを追加
	back_ls_push(link_table_backend, key, (obj["table_id"], obj["rec_id"]))

# 双方向にリンクを追加
def add_sym_link(
	subj,	# 「自分側」の情報 (rec_idとrole_info3つ)
	obj,	# 「相手側」の情報 (rec_idとrole_info3つ)
	link_table_backend,	# link tableの入っているバックエンド
	link_table_id,	# link table ID
):
	# 1-Nチェック (片側; 制約違反の場合は例外を投げる)
	check_1N(subj, obj, link_table_backend, link_table_id)
	check_1N(obj, subj, link_table_backend, link_table_id)	# 1-Nチェック (片側; 制約違反の場合は例外を投げる)
	# 両側のリンクを追加
	add_link(subj, obj, link_table_backend, link_table_id)	# 片側のリンク追加
	add_link(obj, subj, link_table_backend, link_table_id)	# 片側のリンク追加

# rec指定をrec_id指定に統一化 (どちらでもないtypeは例外を投げる)
def to_rec_id(original_query):
	# rec_id指定の場合
	if type(original_query) == type(""): return original_query
	# レコード型の場合
	if original_query.__class__ == Record: return original_query.id
	# 想定外の型の場合
	raise Exception("[SymLinkDB error] invalid type.")

# 自分側のリンクから相手への参照を削除
def delete_link(subj, obj, link_table_backend, link_table_id):
	backend_key = ("link_rec", link_table_id,
		subj["role"], obj["role"],
		subj["table_id"], subj["rec_id"])
	# 元々のレコード一覧を取得 (tableID, recIDの組のリスト)
	org_link_ls = link_table_backend[backend_key]
	# 削除対象のレコード以外を書き戻す
	new_link_ls = []
	for table_id, rec_id in org_link_ls:
		if (table_id == obj["table_id"]) and (rec_id == obj["rec_id"]): continue
		new_link_ls.append((table_id, rec_id))
	link_table_backend[backend_key] = new_link_ls
	# 削除できていない場合に例外を出す
	if len(new_link_ls) == len(org_link_ls):
		raise Exception("[SymLinkDB error] Unable to delete the link. There may be internal data inconsistency.")
	# 空のlink_recになったら、link_recそのものを削除 (注意：linkが関係するレコード自身が削除されたときにゴミが残ること等を避けるため)
	if len(new_link_ls) == 0: del link_table_backend[backend_key]

# 双方向にリンクを削除
def delete_sym_link(
	subj,	# 「自分側」の情報 (rec_idとrole_info3つ)
	obj,	# 「相手側」の情報 (rec_idとrole_info3つ)
	link_table_backend,	# link tableの入っているバックエンド
	link_table_id,	# link table ID
):
	# 自分側のリンクから相手への参照を削除
	delete_link(subj, obj,
		link_table_backend = link_table_backend,
		link_table_id = link_table_id)
	# 相手側のリンクから自分への参照を削除
	delete_link(obj, subj,
		link_table_backend = link_table_backend,
		link_table_id = link_table_id)

# LinkSetクラス [SymLinkDB]
class LinkSet:
	# 初期化処理
	def __init__(self,
		link_table_id,	# link table ID
		link_table_name,	# link table名
		subj_role_info,	# 自roleの情報
		obj_role_info,	# 相手側roleの情報
		rec	# recオブジェクト
	):
		# リンクを特定する情報の登録
		self.link_table_id = link_table_id
		self.link_table_name = link_table_name
		self.subj_role_info = subj_role_info
		self.obj_role_info = obj_role_info
		# 所属するrecやsldbに関する情報の登録
		self.rec = rec
		self.sldb = self.rec.table.sldb
		# link tableの入っているバックエンドの参照を登録
		if self.link_table_name not in self.sldb.loaded_link_tables_dic: raise Exception(f"[SymLinkDB error] specified link table not loaded (link table name = {self.link_table_name})")
		self.link_table_backend = self.sldb.loaded_link_tables_dic[self.link_table_name]
		self.backend_key = ("link_rec", self.link_table_id,
			self.subj_role_info["role"], self.obj_role_info["role"],
			self.subj_role_info["table_id"], self.rec.id)
	# linkの追加 [SymLinkDB]
	def push(self, obj_rec_id):
		if type(obj_rec_id) in [type([]), tuple]:
			for one_id in obj_rec_id: self.push(one_id)
			return None
		# 単一要素をpushする場合
		obj_rec_id = to_rec_id(obj_rec_id)	# rec指定をrec_id指定に統一化 (どちらでもないtypeは例外を投げる)
		obj_table = self.sldb.get_table_by_id(self.obj_role_info["table_id"])	# table_idからTableオブジェクトを引き当て
		if obj_rec_id not in obj_table: raise Exception("[SymLinkDB error] specified record ID does not exist in the table")	# 指定されたrec idが正しくobj tableに存在するかを確認
		if obj_rec_id in self: raise Exception("[SymLinkDB error] This is a link that has already been added")
		# 双方向にリンクを追加
		add_sym_link(
			subj = {"rec_id": self.rec.id, **self.subj_role_info},	# 「自分側」の情報 (rec_idとrole_info3つ)
			obj = {"rec_id": obj_rec_id, **self.obj_role_info},	# 「相手側」の情報 (rec_idとrole_info3つ)
			link_table_backend = self.link_table_backend,	# link tableの入っているバックエンド
			link_table_id = self.link_table_id,	# link table ID
		)
	# linkの存在確認 [SymLinkDB]
	def __contains__(self, obj_rec_id):
		value = (self.obj_role_info["table_id"], to_rec_id(obj_rec_id))	# rec指定をrec_id指定に統一化 (どちらでもないtypeは例外を投げる)
		return (value in self.__get_links())
	# linkの削除 [SymLinkDB]
	def __delitem__(self, obj_rec_id):
		if type(obj_rec_id) in [type([]), tuple]:
			for one_id in obj_rec_id: del self[one_id]
			return None
		# 単一要素をdeleteする場合
		obj_rec_id = to_rec_id(obj_rec_id)	# rec指定をrec_id指定に統一化 (どちらでもないtypeは例外を投げる)
		if obj_rec_id not in self:	# 現在対象としているレコード (subj_rec) が対象としているlink_tableのリンクをそもそも持っていない場合
			raise Exception("[SymLinkDB error] You cannot delete a link that does not exist")
		# 双方向にリンクを削除
		delete_sym_link(
			subj = {"rec_id": self.rec.id, **self.subj_role_info},	# 「自分側」の情報 (rec_idとrole_info3つ)
			obj = {"rec_id": obj_rec_id, **self.obj_role_info},	# 「相手側」の情報 (rec_idとrole_info3つ)
			link_table_backend = self.link_table_backend,	# link tableの入っているバックエンド
			link_table_id = self.link_table_id,	# link table ID
		)
	# link_set内の要素をイテレート [SymLinkDB]
	def __iter__(self):
		obj_table = None
		for table_id, rec_id in self.__get_links():	# 対象のrecが持っているリンク一覧の取得
			if obj_table is None:
				obj_table = self.sldb.get_table_by_id(table_id)	# table_idからTableオブジェクトを引き当て
			yield obj_table[rec_id]
	# link_set内の要素数を取得 [SymLinkDB]
	def __len__(self): return len(self.__get_links())	# 対象のrecが持っているリンク一覧の取得
	# 対象のrecが持っているリンク一覧の取得
	def __get_links(self):
		if self.backend_key not in self.link_table_backend: return []
		return self.link_table_backend[self.backend_key]
	# 文字列化 (その1, その2)
	def __str__(self):
		subj_t_name = self.sldb.get_table_by_id(self.subj_role_info['table_id']).table_name	# table_idからTableオブジェクトを引き当て
		obj_t_name = self.sldb.get_table_by_id(self.obj_role_info['table_id']).table_name	# table_idからTableオブジェクトを引き当て
		return (
			f"<LinkSet {self.link_table_name}: " +
			f"{subj_t_name}({self.subj_role_info['role']};{self.subj_role_info['1-N']}) -> " +
			f"{obj_t_name}({self.obj_role_info['role']};{self.obj_role_info['1-N']};n={len(self)})>"
		)
	def __repr__(self): return str(self)

# レコードクラス [SymLinkDB]
class Record:
	# 初期化処理
	def __init__(self, rec_id, table):
		# レコードに関する情報の登録
		self.id = rec_id
		# 所属するtableに関する情報の登録
		self.table = table
		self.table_name = table.table_name
		self.table_id = table.table_id
	# dataのgetter
	@property
	def data(self):
		self.is_deleted(raise_err = True)	# 自身が削除済みかどうかを判定
		return self.table.backend[("rec", self.table_id, self.id)]["data"]
	# dataのsetter
	@data.setter
	def data(self, new_data):
		self.is_deleted(raise_err = True)	# 自身が削除済みかどうかを判定
		# dataが正しく検索キーを持っているか確認する
		if self.table.search_keys is not None: check_data_format(new_data, self.table.search_keys)
		# バックエンドへの書き込み
		old_data = self.table.backend[("rec", self.table_id, self.id)]["data"]
		def add_data(v):
			v["data"] = new_data
			return v
		dic_like_edit(	# 辞書-likeオブジェクトの所定keyのvalueを編集
			self.table.backend, key = ("rec", self.table_id, self.id),
			edit_func = add_data)
		# 検索インデックス関連
		if self.table.search_keys is not None:
			# 検索インデックスの削除
			del_search_idx(old_data, self.table.search_keys, self.table_id, self.id, self.table.backend)
			# 検索インデックスの追加
			add_search_idx(new_data, self.table.search_keys, self.table_id, self.id, self.table.backend)
	# link_setの取得 [SymLinkDB]
	def __getitem__(self, query):
		self.is_deleted(raise_err = True)	# 自身が削除済みかどうかを判定
		# role名あるいはrole名とlink table名からlink情報を取得
		link_info = get_link_info(query, self.table_id, self.table.sldb.backend)
		# LinkSetオブジェクトを作成して返す
		return LinkSet(	# LinkSetクラス [SymLinkDB]
			link_info["link_table_id"],	# link table ID
			link_info["link_table_name"],	# link table名
			subj_role_info = link_info["subj_role_info"],	# 自roleの情報
			obj_role_info = link_info["obj_role_info"],	# 相手側roleの情報
			rec = self)	# linkオブジェクト
	# 文字列化 (その1, その2)
	def __str__(self):
		if self.is_deleted() is True: return "<SymLinkRecord (deleted)>"	# 自身が削除済みかどうかを判定
		links = souts(list(self), None)
		return f"<SymLinkRecord {self.table_name}.{self.id} data = {souts(self.data)} links = {links}>"
	def __repr__(self): return str(self)
	# link一覧をイテレートして取得 [SymLinkDB]
	def __iter__(self):
		self.is_deleted(raise_err = True)	# 自身が削除済みかどうかを判定
		# link_tableのentityのみに絞り込み
		sys_backend = self.table.sldb.backend
		lt_entities = entity_filter(self.table.sldb.backend[("meta", "system_contain_data")],
			{"entity_type": "SLDB-link-table", "sys_backend_id": sys_backend[("meta", "backend_id")]})
		# 全linkを確認し、自recに関係するものをiterする (両側確認)
		for entity in lt_entities:
			link_table_name = entity["link_table_name"]
			for role_dic in entity["rel_table_info"]:
				key = (link_table_name, role_dic["role"])
				if key in self: yield key
	# 指定されたlinkの存在確認 [SymLinkDB]
	def __contains__(self, query):
		self.is_deleted(raise_err = True)	# 自身が削除済みかどうかを判定
		try:
			self[query]
			return True
		except:
			return False
	# 自身が削除済みかどうかを判定
	def is_deleted(self, raise_err = False):
		deleted = (self.id not in self.table)
		if deleted is True and raise_err is True: raise Exception("[SymLinkDB error] This record is already deleted.")
		return deleted

# dataが正しく検索キーを持っているか確認する
def check_data_format(data, search_keys):
	# 辞書形式でない場合はエラー
	if type(data) != type({}):
		raise Exception("[SymLinkDB error] Insert data in dictionary format into the table where the search_key is set.")
	# search_keyを1つでも持たない場合はエラー
	for key in search_keys:
		if key not in data: 
			raise Exception("[SymLinkDB error] Specified search_key is not included.")
	return True

# 検索対象データをjson文字列にする
def to_json_str(arg_data):
	try:
		return json.dumps(arg_data, sort_keys=True)
	except TypeError as e:
		raise Exception(f"[SymLinkDB error] Value of the search_key must be JSON-able\n{e}")

# 検索インデックスの追加
def add_search_idx(data, search_keys, table_id, rec_id, backend):
	for key in search_keys:
		# 該当keyの値をjson文字列にする
		val_str = to_json_str(data[key])	# 検索対象データをjson文字列にする
		# 値が既にインデックスとして初期化されていない場合は初期化
		idx_key = ("search_idx", table_id, key, val_str)
		if idx_key not in backend: backend[idx_key] = {}
		# バックエンドへの追加
		def add_key(v):
			v[rec_id] = True
			return v
		dic_like_edit(	# 辞書-likeオブジェクトの所定keyのvalueを編集
			backend, key = idx_key,
			edit_func = add_key)

# 検索インデックスの削除
def del_search_idx(data, search_keys, table_id, rec_id, backend):
	for key in search_keys:
		# 該当keyの値をjson文字列にする
		val_str = to_json_str(data[key])	# 検索対象データをjson文字列にする
		# バックエンドからの削除
		def del_key(v):
			del v[rec_id]
			return v
		dic_like_edit(	# 辞書-likeオブジェクトの所定keyのvalueを編集
			backend, key = ("search_idx", table_id, key, val_str),
			edit_func = del_key)

# リストの共通要素を見つける
def find_common_elements(id_dic_ls):
	if len(id_dic_ls) == 0: raise Exceprion("[SymLinkDB error] Unexpected error in the program.")
	# 最初の辞書を共通要素の候補とする
	common_elements = set(id_dic_ls[0])
	# 残りのリストに対して共通要素を探す
	for id_dic in id_dic_ls[1:]:
	    common_elements.intersection_update(id_dic)
	# 共通要素をリストとして返す
	return list(common_elements)

# Tableクラス
class Table:
	# 初期化処理 (table自体を作る処理はすでに実行されている想定)
	def __init__(self,
		table_name,	# table名
		table_id,	# table-ID
		search_keys, # 検索用に使える列名
		backend,	# tableのデータが存在するbackend
		sldb	# 所属するsldbオブジェクト
	):
		self.table_name = table_name
		self.table_id = table_id
		self.search_keys = search_keys
		self.backend = backend
		self.sldb = sldb
	# レコードの新規作成 [SymLinkDB]
	def create(self, data, links = None):
		# dataが正しく検索キーを持っているか確認する
		if self.search_keys is not None: check_data_format(data, self.search_keys)
		# linksが省略された場合
		if links is None: links = {}
		# rec_idの発行
		rec_id = slim_id.gen(lambda arg_id: (arg_id in self))
		# レコードデータをbackendに書き込み
		self.backend[("rec", self.table_id, rec_id)] = {"data": data}
		# id一覧に書き込み
		def add_key(v):
			v[rec_id] = True
			return v
		dic_like_edit(	# 辞書-likeオブジェクトの所定keyのvalueを編集
			self.backend, key = ("table_info", "all_rec_ids", self.table_id),
			edit_func = add_key)
		# レコード数の更新
		dic_like_edit(	# 辞書-likeオブジェクトの所定keyのvalueを編集
			self.backend, key = ("table_info", "rec_n", self.table_id),
			edit_func = lambda v: v + 1)
		# 検索インデックスの追加
		if self.search_keys is not None:
			add_search_idx(data, self.search_keys, self.table_id, rec_id, self.backend)
		# link情報の書き込み
		new_rec = self[rec_id]	# 新しく生成したレコード
		for query in links:
			if links[query] is None: continue
			new_rec[query].push(links[query])
		# 新規生成したrec_idを返す
		return self[rec_id]
	# レコードの取得 [SymLinkDB]
	def __getitem__(self, rec_id):
		# 検索条件での指定
		if type(rec_id) == type({}):
			rec_ls = self.search_rec(search_cond_dic = rec_id)	# 検索条件でレコードIDを取得
			return rec_ls
		# 単一要素を処理する場合
		rec_id = to_rec_id(rec_id)	# rec指定をrec_id指定に統一化 (どちらでもないtypeは例外を投げる)
		return Record(rec_id, table = self)	# レコードクラス [SymLinkDB]
	# レコードの存在確認 [SymLinkDB]
	def __contains__(self, rec_id):
		rec_id = to_rec_id(rec_id)	# rec指定をrec_id指定に統一化 (どちらでもないtypeは例外を投げる)
		all_rec_ids = self.backend[("table_info", "all_rec_ids", self.table_id)]
		return (rec_id in all_rec_ids)
	# レコード数取得 [SymLinkDB]
	def __len__(self): return self.backend[("table_info", "rec_n", self.table_id)]
	# tableのイテレート [SymLinkDB]
	def __iter__(self):
		for rec_id in self.backend[("table_info", "all_rec_ids", self.table_id)]:
			yield self[rec_id]
	# レコードの削除 [SymLinkDB]
	def __delitem__(self, rec_id):
		rec_id = to_rec_id(rec_id)	# rec指定をrec_id指定に統一化 (どちらでもないtypeは例外を投げる)
		if rec_id not in self: raise Exception("[SymLinkDB error] You cannot delete a record that does not exist")
		# 削除対象が関係するリンクの削除
		rec = self[rec_id]
		for link_query in rec:
			link_set = rec[link_query]
			for obs_rec in link_set: del link_set[obs_rec]
		# 検索インデックスの削除
		if self.search_keys is not None:
			del_search_idx(rec.data, self.search_keys, self.table_id, rec_id, self.backend)
		# レコード記録本体の削除
		del self.backend[("rec", self.table_id, rec_id)]
		# id一覧から削除
		def del_key(v):
			del v[rec_id]
			return v
		dic_like_edit(	# 辞書-likeオブジェクトの所定keyのvalueを編集
			self.backend, key = ("table_info", "all_rec_ids", self.table_id),
			edit_func = del_key)
		# レコード数の更新 (デクリメント)
		dic_like_edit(	# 辞書-likeオブジェクトの所定keyのvalueを編集
			self.backend, key = ("table_info", "rec_n", self.table_id),
			edit_func = lambda v: v - 1)
	# 検索条件でレコードIDを取得
	def search_rec(self, search_cond_dic):
		if len(search_cond_dic) == 0: raise Exception("[SymLinkDB error] Please specify one or more keys for the search condition.")
		cand_ls = []
		for search_key in search_cond_dic:
			if search_key not in self.search_keys: raise Exception("[SymLinkDB error] You cannot search by a key that is not specified as the search_key during load_table.")
			# 該当keyの値をjson文字列にする
			search_val_str = to_json_str(search_cond_dic[search_key])	# 検索対象データをjson文字列にする
			# rec_id候補を取得する
			idx_key = ("search_idx", self.table_id, search_key, search_val_str)
			if idx_key not in self.backend: return []	# 値がそもそもインデックスとして初期化されていない場合
			cand_ls.append(self.backend[idx_key])
		# すべての検索条件に合致するrec_idに絞り込む
		rec_id_ls = find_common_elements(cand_ls)	# リストの共通要素を見つける
		# recを取得して返す
		return [self[rec_id] for rec_id in rec_id_ls]	# レコードの取得 [SymLinkDB]
	# 文字列化 (その1, その2)
	def __str__(self):
		recs_str = gen_recs_str(self)	# レコード一覧を表した要約文字列の生成
		return f"<SymLinkTable {self.table_name} recs = {recs_str} search_keys = {souts(self.search_keys, 3)} >"
	def __repr__(self): return str(self)

# versionの書き換え
def rewrite_data_version(backend, target_key, filter_condition, new_ver_str):
	old_value = backend[target_key]
	def rewriter(entity):
		entity["data_version"] = new_ver_str
		return entity
	def judge_func(entity):
		for k in filter_condition:
			if k not in entity: return False
			if entity[k] != filter_condition[k]: return False
		return True
	new_value = [
		(rewriter(e) if judge_func(e) else e)
		for e in old_value]
	backend[target_key] = new_value

# keyが旧形式の場合にデータを修正する (同一table内リンクにおいてリンクが両方向に増殖するバグに関連)
def link_dup_bug_fix(
	filter_condition,	# エンティティのフィルタリング条件 (uniqueになるようなもの)
	sys_backend,	# DB全体情報 (system) を保管するバックエンド
	table_data_backend,	# テーブルのデータを保管するバックエンド
):
	# バージョンを確認 (修正要否を判断)
	entity = entity_filter(sys_backend[("meta", "system_contain_data")],	# link_tableのentityのみに絞り込み
		filter_condition, unique = True)
	data_version = entity.get("data_version", "with_same_table_link_duplication_bug_ver")
	if data_version != "with_same_table_link_duplication_bug_ver": return "no_fix_needed"
	# 修正
	keys = list(table_data_backend)
	for key in keys:
		# 対象の絞り込み (link_table_idが一致するlink_rec)
		if type(key) != tuple: continue
		if key[:2] != ("link_rec", entity["link_table_id"]): continue
		# データ形式が想定外の場合
		if len(key) != 4: raise Exception("unknown error has occurred.")
		# 新形式のkeyに書き換え
		value = table_data_backend[key]
		del table_data_backend[key]
		info_A, info_B = entity["rel_table_info"]
		for subj, obj in [(info_A, info_B), (info_B, info_A)]:
			if key[2] != subj["table_id"]: continue
			new_key = key[:2] + (subj["role"], obj["role"]) + key[2:]
			table_data_backend[new_key] = value
	# versionの書き換え
	rewrite_data_version(sys_backend, ("meta", "system_contain_data"), filter_condition, "same_table_link_duplication_bug_fix_ver")
	rewrite_data_version(table_data_backend, ("meta", "backend_contain_data"), filter_condition, "same_table_link_duplication_bug_fix_ver")	# versionの書き換え
	# コミット
	sys_backend.commit()
	table_data_backend.commit()
	return "fixed"

# linkテーブル新規作成処理
def create_link_table(
	rel_info0,	# 関係性情報0 (table名, role, 1 or N)
	rel_info1,	# 関係性情報1 (table名, role, 1 or N)
	link_table_name,	# テーブル名
	sys_backend,	# DB全体情報 (system) が保管されているバックエンド
	link_table_data_backend,	# linkテーブルのデータが保管されているバックエンド
	sldb	# sldbオブジェクト
):
	# role名規則チェック
	if rel_info0[1] == rel_info1[1]: raise Exception("[SymLinkDB error] The role names at both ends of the link must be different.")
	# link table初期化不整合チェック
	filtered_entities = [
		e for e in link_table_data_backend[("meta", "backend_contain_data")]
		if (
			e.get("link_table_name") == link_table_name and
			e.get("sys_backend_id") == sys_backend[("meta", "backend_id")]
		)
	]
	if len(filtered_entities) > 0: raise Exception("[SymLinkDB error] Inconsistency in the stored link table information. According to the information on the DB-system side, the link table that is currently being created is uninitialized, but within the backend that stores the link table, it is already initialized. Degradation in material management is suspected.")
	# link table情報
	link_table_data = {
		"entity_type": "SLDB-link-table",
		"data_version": "same_table_link_duplication_bug_fix_ver",	# データ形式
		"link_table_id": gen_unique_id(),	# 十分な長さの64進idの発行 (uuidの桁数を参考に決定)
		"sys_backend_id": sys_backend[("meta", "backend_id")],
		"data_backend_id": link_table_data_backend[("meta", "backend_id")],
		"link_table_name": link_table_name,
		"rel_table_info": [
			{"table_id": sldb[rel_info0[0]].table_id, "role": rel_info0[1], "1-N": rel_info0[2]},
			{"table_id": sldb[rel_info1[0]].table_id, "role": rel_info1[1], "1-N": rel_info1[2]},
		]
	}
	# systemにtable情報を書き込む
	back_ls_push(	# 辞書-likeオブジェクトの所定のkeyのvalue(list)にデータを追加
		sys_backend, ("meta", "system_contain_data"),
		link_table_data)
	# backendにtable情報を書き込む
	back_ls_push(	# 辞書-likeオブジェクトの所定のkeyのvalue(list)にデータを追加
		link_table_data_backend, ("meta", "backend_contain_data"),
		link_table_data)

# SLDBクラスのオブジェクトを返す
class SLDB:
	# 初期化処理
	def __init__(self, backend):
		self.backend = backend
		# SLDB-backendとして初期化 (必要な場合)
		if ("meta", "backend_id") not in self.backend:
			init_sldb_backend(self.backend)		# SLDB-backendとして初期化
		# さらにDB-systemとして初期化 (必要な場合)
		if judge_sys_initialized(self.backend) is not True:	# このバックエンドがSLDB-sysとして初期化されているかを判定
			init_sldb_sys(self.backend)	# このバックエンドをSLDB-sysとして初期化
		# loadが完了しているtable一覧 (valueはTableクラス)
		self.loaded_tables_dic = {}
		# loadが完了しているlink table一覧 (valueはbackend)
		self.loaded_link_tables_dic = {}
	# table数取得 [SymLinkDB]
	def __len__(self):
		sc_data = self.backend[("meta", "system_contain_data")]
		return len([e for e in sc_data
			if e["entity_type"] == "SLDB-table"])
	# tableの存在確認 [SymLinkDB]
	def __contains__(self, table_name):
		for t_name in self:
			if t_name == table_name: return True
		return False
	# tableのイテレート [SymLinkDB]
	def __iter__(self):
		sc_data = self.backend[("meta", "system_contain_data")]
		for e in sc_data:
			if e["entity_type"] != "SLDB-table": continue
			if e["sys_backend_id"] != self.backend[("meta", "backend_id")]: continue
			yield e["table_name"]
	# tableの読み込み (存在しない場合は空で初期化される) [SymLinkDB]
	def load_table(self, table_name, search_keys = None, backend = None):
		sys_backend = self.backend
		if backend is None: backend = sys_backend
		table_filter = {"sys_backend_id": sys_backend[("meta", "backend_id")], "table_name": table_name}	# 対象テーブルを一意に特定する条件 (DBをまたぐとtable_nameが衝突する可能性があることに留意すべし)
		if table_name in self:	# tableの存在確認 [SymLinkDB]
			# テーブルがすでにある場合
			check_entity_backend(	# あるエンティティについて、バックエンドが整合しているかを確かめる (NGの場合は例外を送出する)
				filter_condition = table_filter,	# エンティティのフィルタリング条件 (uniqueになるようなもの)
				sys_backend = sys_backend,	# DB全体情報 (system) を保管するバックエンド
				table_data_backend = backend,	# テーブルのデータを保管するバックエンド
			)
			# search_keysの整合性を確認
			check_search_keys(
				search_keys,	# チェック対象のsearch_keys
				table_name,	# テーブル名
				sys_backend[("meta", "system_contain_data")],	# 該当sysの内容物一覧
			)
		else:
			# テーブルがない場合
			if ("meta", "backend_id") not in backend:	# SLDB-backendとして初期化 (必要な場合)
				init_sldb_backend(backend)		# SLDB-backendとして初期化
			create_table(	# テーブル新規作成処理
				table_name = table_name,	# テーブル名
				search_keys = search_keys,	# 検索用に使いたい列名
				sys_backend = sys_backend,	# DB全体情報 (system) が保管されているバックエンド
				table_data_backend = backend,	# テーブルのデータが保管されているバックエンド
			)
		# load済み一覧に追加
		table_id = entity_filter(sys_backend[("meta", "system_contain_data")], table_filter, unique = True)["table_id"]	# metaのcontain_dataを条件でフィルタリングする
		self.loaded_tables_dic[table_name] = Table(table_name, table_id, search_keys, backend, self)	# Tableクラス
	# Tableの取得
	def __getitem__(self, table_name):
		if table_name in self.loaded_tables_dic:
			return self.loaded_tables_dic[table_name]
		# 2種類の非存在エラー
		if table_name in self: raise Exception("[SymLinkDB error] The specified table exists, but has not been loaded. Please load it using the load_table() function.")
		raise Exception("[SymLinkDB error] The specified table does not exist.")
	# linkの読み込み (存在しない場合は空で初期化される) [SymLinkDB]
	def load_link(self,
		link_table_name,	# テーブル名
		rel_info0,	# 関係性情報0 (table名, role, 1 or N)
		rel_info1,	# 関係性情報1 (table名, role, 1 or N)
		backend = None	# バックエンド (None指定でsystemのデフォルトバックエンド)
	):
		if backend is None: backend = self.backend
		if self.__contain_link(link_table_name) is True:	# linkの存在確認 [SymLinkDB]
			# linkテーブルがすでにある場合
			filter_condition = {"sys_backend_id": self.backend[("meta", "backend_id")],	# link_tableを一意に特定する条件
				"link_table_name": link_table_name}
			check_entity_backend(	# あるエンティティについて、バックエンドが整合しているかを確かめる (NGの場合は例外を送出する)
				filter_condition = filter_condition,	# エンティティのフィルタリング条件 (uniqueになるようなもの)
				sys_backend = self.backend,	# DB全体情報 (system) を保管するバックエンド
				table_data_backend = backend,	# テーブルのデータを保管するバックエンド
			)
			link_dup_bug_fix(	# keyが旧形式の場合にデータを修正する (同一table内リンクにおいてリンクが両方向に増殖するバグに関連)
				filter_condition = filter_condition,	# エンティティのフィルタリング条件 (uniqueになるようなもの)
				sys_backend = self.backend,	# DB全体情報 (system) を保管するバックエンド
				table_data_backend = backend,	# テーブルのデータを保管するバックエンド
			)
		else:
			# linkテーブルがない場合
			if ("meta", "backend_id") not in backend:	# SLDB-backendとして初期化 (必要な場合)
				init_sldb_backend(backend)		# SLDB-backendとして初期化
			create_link_table(	# linkテーブル新規作成処理
				rel_info0,	# 関係性情報0 (table名, role, 1 or N)
				rel_info1,	# 関係性情報1 (table名, role, 1 or N)
				link_table_name = link_table_name,	# テーブル名
				sys_backend = self.backend,	# DB全体情報 (system) が保管されているバックエンド
				link_table_data_backend = backend,	# linkテーブルのデータが保管されているバックエンド
				sldb = self	# sldbオブジェクト
			)
		# load済み一覧に追加
		self.loaded_link_tables_dic[link_table_name] = backend
	# linkの存在確認 [SymLinkDB]
	def __contain_link(self, link_table_name):
		sc_data = self.backend[("meta", "system_contain_data")]
		for e in sc_data:
			if e["entity_type"] != "SLDB-link-table": continue
			if e["sys_backend_id"] != self.backend[("meta", "backend_id")]: continue
			if e["link_table_name"] == link_table_name: return True
		return False
	# 強制コミット [SymLinkDB]
	def commit(self):
		# 関係するbackendを列挙
		backend_dic = {self.backend[("meta", "backend_id")]: self.backend}	# systemのbackend
		for table_name in self:	# 所属している全tableのbackendを列挙
			backend = self[table_name].backend
			backend_dic[backend[("meta", "backend_id")]] = backend
		# 列挙したbackendをcommit
		for backend_id in backend_dic:
			backend_dic[backend_id].commit()
	# 強制コミット [SymLinkDB]
	def commit(self):
		# backendを一時登録
		backend_dic = {}
		def reg(backend): backend_dic[backend[("meta", "backend_id")]] = backend
		# 関係するbackendを列挙
		reg(self.backend)	# system
		for name in self.loaded_tables_dic: reg(self.loaded_tables_dic[name].backend)	# tables
		for name in self.loaded_link_tables_dic: reg(self.loaded_link_tables_dic[name])	# link-tables
		# 列挙したbackendをcommit
		for backend_id in backend_dic:
			backend_dic[backend_id].commit()
	# 文字列化 (その1, その2)
	def __str__(self):
		table_name_ls = list(self)	# table名一覧を取得
		return f"<SymLinkDB tables = {souts(table_name_ls, 3)}>"
	def __repr__(self): return str(self)
	# table_idからTableオブジェクトを引き当て
	def get_table_by_id(self, table_id):
		entity = entity_filter(	# metaのcontain_dataを条件でフィルタリングする
			self.backend[("meta", "system_contain_data")],
			{"table_id": table_id}, unique = True)
		return self[entity["table_name"]]

# SLDB初期化 [SymLinkDB]
def conn(backend):
	# SLDBクラスのオブジェクトを返す
	sldb = SLDB(backend)
	# sldbをプログラム終了時自動commit対象に追加
	commit_target_dbs.append(sldb)
	return sldb




# CachedFileDic-backend (SLDB-backend) [cfd_backend]
# 【動作確認 / 使用例】

import sys
from sout import sout
from relpath import add_import_path
add_import_path("../")
# CachedFileDic-backend (SLDB-backend) [cfd_backend]
import cfd_backend

# CachedFileDicバックエンドへの接続 [cfd_backend]
cfd_back = cfd_backend.conn("./test_db/")

cfd_back["key_1"] = "hoge"	# create, update共通 [cfd_backend]
print(cfd_back["key_1"])	# key読み出し [cfd_backend]
for key in cfd_back: print(key)	# iter (for文脈等での利用) [memory_backend]
del cfd_back["key_1"]	# 削除 [cfd_backend]
print("key_1" in cfd_back)	# 存在確認 [cfd_backend]
cfd_back.commit()	# 強制commit [cfd_backend]

# (非公式インターフェース) 文字列化
print(cfd_back)




# 単純メモリbackend (SLDB-backend) [memory_backend]
# 【動作確認 / 使用例】

import sys
from sout import sout
from relpath import add_import_path
add_import_path("../")
# 単純メモリbackend (SLDB-backend) [memory_backend]
import memory_backend

# メモリバックエンドへの接続 [memory_backend]
m_back = memory_backend.conn("./test_db/")
print("fuga")

m_back["key_1"] = "hoge"	# create, update共通 [memory_backend]
print(m_back["key_1"])	# key読み出し [memory_backend]
del m_back["key_1"]	# 削除 [memory_backend]
print("key_1" in m_back)	# 存在確認 [memory_backend]
m_back.commit()	# 強制commit [memory_backend]

# (非公式インターフェース) 文字列化
print(m_back)




# CachedFileDic-backend (SLDB-backend) [cfd_backend]

import os
import sys
import fies
import atexit
import pickle
import slim_id
import CachedFileDic
from sout import sout, souts
from relpath import add_import_path
add_import_path("../")
# 単純メモリbackend (SLDB-backend) [memory_backend]
import memory_backend

# プログラム終了時にcommitを実行する
commit_target_ls = []	# コミット対象
def cleanup():
	for one_backend in commit_target_ls:
		one_backend.commit()
# プログラム終了時に呼び出す関数を登録
atexit.register(cleanup)

# CachedFileDic-backend (SLDB-backend) [cfd_backend]
class CFD_Backend:
	# 初期化処理
	def __init__(self, data_dir,
		fmt = "fpkl"	# バックエンドのファイル形式 (fpkl: fast-pickle(default), pickle: pickle)
	):
		# インデックスとデータを格納する2つのディレクトリ
		mem_idx_dir = os.path.join(data_dir, "mem_idx")
		cfd_data_dir = os.path.join(data_dir, "cfd_data")
		if not os.path.exists(mem_idx_dir): os.makedirs(mem_idx_dir)
		if not os.path.exists(cfd_data_dir): os.makedirs(cfd_data_dir)
		# インデックスを格納するmemory_backendの初期化
		self.mem_idx = memory_backend.conn(mem_idx_dir, fmt = fmt)	# メモリバックエンドへの接続 [memory_backend]
		# データを格納するCachedFileDicの初期化
		self.cfd_data = CachedFileDic.conn(cfd_data_dir, fmt = fmt)	# 対象ディレクトリに接続 [CachedFileDic]
		# プログラム終了時のcommit対象に自身を追加
		commit_target_ls.append(self)
	# create, update共通 [memory_backend]
	def __setitem__(self, key, value):
		# 新しいinner_keyの生成
		inner_key = slim_id.gen(lambda e: (e in self.cfd_data), length = 22)	# 十分に長いkey
		# 両方の辞書に登録
		self.mem_idx[key] = inner_key
		self.cfd_data[inner_key] = value
	# key読み出し [memory_backend]
	def __getitem__(self, key):
		inner_key = self.mem_idx[key]
		return self.cfd_data[inner_key]
	# 削除 [memory_backend]
	def __delitem__(self, key): del self.mem_idx[key]	# メモリーバックエンドのみから削除
	# 存在確認 [memory_backend]
	def __contains__(self, key): return (key in self.mem_idx)
	# iter (for文脈等での利用) [memory_backend]
	def __iter__(self): return iter(self.mem_idx)
	# 強制commit [memory_backend]
	def commit(self):
		self.mem_idx.commit()
		self.cfd_data.commit()
	# 文字列化 (その1, その2)
	def __str__(self):
		# memory-backendの非公式APIを利用
		for key_example in self.mem_idx: break	# keyの例を取得
		size = len(self.mem_idx)
		data_str = ("" if size == 0
			else f'{key_example}: {souts(self[key_example])}' +
				("" if size == 1 else f", ...(n={size})")
		)
		return f"<SLDB-CFD_backend {{{data_str}}}>"
	def __repr__(self): return str(self)

# CachedFileDicバックエンドへの接続 [cfd_backend]
def conn(
	data_dir,	# バックエンドの情報を記録するディレクトリ
	fmt = "fpkl"	# バックエンドのファイル形式 (fpkl: fast-pickle(default), pickle: pickle)
):
	# バックエンドのクラス
	return CFD_Backend(data_dir, fmt = fmt)




from setuptools import setup
# 公開用パッケージの作成 [ezpip]
import ezpip

# 公開用パッケージの作成 [ezpip]
with ezpip.packager(develop_dir = "./_develop_SymLinkDB/") as p:
	setup(
		name = "SymLinkDB",
		version = "0.4.4",
		description = "A graph database characterized by bidirectional links. It allows for intuitive design of data relationships.",
		author = "bib_inf",
		author_email = "contact.bibinf@gmail.com",
		url = "https://github.co.jp/",
		packages = p.packages,
		install_requires = ["ezpip", "sout>=1.2.1", "relpath", "slim-id>=0.0.2", "fies>=1.4.0", "CachedFileDic>=0.2.0"],
		long_description = p.long_description,
		long_description_content_type = "text/markdown",
		license = "CC0 v1.0",
		classifiers = [
			"Programming Language :: Python :: 3",
			"License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication"
		],
	)
